/*------------------------------------------------------------------------------
                               XFR
                    By Donald G. Krantz 3/84
               (C) Copyright 1984 Donald G. Krantz
                       All Rights Reserved

XFR is a packet protocol file transferer similar to modem7. It is
designed for 19.2K baud transfers "nose-to-nose" between computers
and works very well in modem use at low speeds. XFR is compatible
with most versions of Modem7.

08/13/87 -- adapted to Megamax C compiler for the Atari ST by J.K.LaPeer

------------------------------------------------------------------------------*/

#include <stdio.h>		/* ECO C */
#include <osbind.h>

#define FALSE	0
#define TRUE	1

/*---------------------------------------------------------------
ASCII and XMODEM control characters
---------------------------------------------------------------*/

#define SOH 1			/* start of header		*/
#define EOT 4			/* end of transmission		*/
#define ACK 6			/* true acknowledge		*/
#define NAK 0x15		/* false acknowledge		*/
#define CRC 'C'			/* request CRC mode		*/
#define CAN 0x18		/* cancel transmission		*/
#define EoF 26			/* end of file (used for name)	*/
#define BADNAME 0x75		/* received bad name checksum	*/
#define BELL 7			/* console bell			*/
#define NEWLINE 10		/* linefeed char		*/
#define CR 13			/* carriage return char		*/
#define BS 8			/* backspace character		*/
#define ESCTERM 'E'-'@'		/* escape from terminal mode	*/

/*---------------------------------------------------------------
User accessable system equates
---------------------------------------------------------------*/

#define RETRY 5			/* no. of retrys before abort	*/
#define RECSIZE 128		/* transfer record size		*/
#define NAMESIZE 11		/* filename fixed length 	*/ 
#define TITLE printf( "\t\t\t  XFR - File Transfer Utility\n")
#define VERSION printf("\t\t\t ZORBA Version 2.00 as of 3/85\n")
#define CLS printf( "\033E")	/* clear screen			*/
#define ERROR -1
#define MAGIC_NUMBER 10000	/* time constant - machine dep.	*/
#define VOID int		/* function type		*/
#define LINK		/ * file rename supported	*/

/*---------------------------------------------------------------
Machine dependant port addresses - ZORBA Version (Intel 8251A)
---------------------------------------------------------------*/

#define RXDATA 0x20		/* rec'd data UART port		*/
#define TXDATA 0x20		/* tx'ed data UART port		*/
#define UARTCMD 0x21		/* Command/Status port		*/

/*---------------------------------------------------------------
Global variables
---------------------------------------------------------------*/

char *line;			/* scratch input line 		*/
char *buffer;			/* record buffer		*/
FILE *fd;			/* file POINTER!		*/
unsigned rec;			/* record number		*/
char checksum;			/* checksum accumulator		*/
unsigned crcaccum;		/* global crc bytes		*/
char crc;			/* CRC on/off flag		*/
char batch;			/* batch on/off flag		*/
jmp_buf to_menu;		/* long jump envelope		*/
struct {			/* batch list array		*/
         long size_part;
         int  time_part, date_part;
         char name_part[ 65 ];
	} list[ 128 ];

/*----------------------------------------------------------------
Function type declarations
----------------------------------------------------------------*/

VOID main(), term(), transmit(), trans(), txname(), txrec();
VOID receive(), rxname(), rxfile(), make(), error(), clrcrc();
VOID updcrc(), abort(), sleep(), quit(), init(), tx(), rxbatch();
char *index(), *parse(), *unparse();
int fillbuf(), rxstat(), txstat();
unsigned swait(), wait(), waitcan(), rx();

/*---------------------------------------------------------------
Executable code follows.

main() is (of course) the master program loop. Uses setjmp()
to un-nest function calls on error exits.

References globals: crc, tty;
Modifies globals: crc, tty;
---------------------------------------------------------------*/

int GetChar()
{
  int	ch;
  
  ch = (int)Bconin(2);
  Bconout(2, ch);

  return (ch);

}
/*
putchar(c)
int c;
{
  Bconout(2, c);
}

dump(Block, Width)
char  *Block;
int   Width;
{
  char  Ascii[16];
  int   i;
  
  if (Width) {
          i = 0;
          while(Width--) {
               if ((*Block >= ' ') && (*Block <= 0x7e))
                       Ascii[i++] = *Block;
                  else Ascii[i++] = '.';
               printf("%02x ", *Block++);
               if (i == 16) {
                       Ascii[i] = '\0';
                       printf(" %s\n", Ascii);
                       i = 0;
               }
          }
          if (i > 0) {
                  Ascii[i] = '\0';
                  printf(" %s\n", Ascii);
          }
  }

}
*/
VOID main(argc, argv)
int	argc;
char	*argv[];
{
   char menu;

	menu = FALSE;		/* initially expert mode	*/
	init();			/* initialize modem and UART	*/
	TITLE;
	VERSION;
	printf( "Type 'M' for MENU display....\n");
	setjmp( &to_menu );	/* menu loop			*/
	while( TRUE )
	{
		printf( "\nMode: %s\n", crc ? "CRC" : "Checksum" );
		if( menu )
		{
		  printf( "\n\tS Send file    R Receive file\n" );
		  printf( "\tT FDX term     V CRC/Checksum\n" );
		  printf( "\tM toggle menu  X Exit\n" );
		}
		printf( "\nCommand: " );
		switch( toupper( GetChar() ) )
		{
			case 'S':		/* send file(s)	*/
				transmit();
				break;
			case 'R':		/* RX file(s)	*/
				receive();
				break;
			case 'T':		/* dumb terminal*/
				term();
				break;
			case 'V':		/* toggle crc	*/
				crc = !crc;
				break;
			case 'M':		/* toggle menu	*/
				menu = !menu;
				break;
			case 'X':		/* quit		*/
				quit();
			default:		/* option bad	*/
				putchar( BELL );
		}
	}
}


/*---------------------------------------------------------------
term() is the dumb terminal section. Escape by typing ESCTERM
character (see ASCII #defines). 
Some machine dependent stuff.
---------------------------------------------------------------*/

VOID term()
{
   char ch;				/* scratch char		*/

	CLS;
	while( TRUE )			/* the loop.		*/
	{
		if( rxstat() )		/* if char coming in	*/
			putchar( rx() );/* display local	*/
		if( Bconstat(2) != 0 ) {
		    ch = (char)Bconin(2);
		    if( ch == ESCTERM ) return;  /* check exit  */
			else tx( ch );	/* disp remote		*/
	        }
	}
}


/*---------------------------------------------------------------
transmit() is the transmission executive. It requires no input
parameters. This function takes filename(s) as input from the
user, one line at a time. A comma at the end of a filename
signals the intention to enter anther filename on the next
line. More than one filename entered sets "batch" mode. 32
names maximum, for no especially good reason.

Modifies globals: buffer, batch;
---------------------------------------------------------------*/

VOID transmit()
{
  int     ct = 0;
  int     i;
  int     done;
  char    TempName[65];
  struct  { char     filler[21];
            char     fattr;
            unsigned ftime;
            unsigned fdate;
            long     fsize;
            char     fname[15];
            char     filler2[32]; /* a wastefull insurance */
          } SysDta;
          
	printf( "\n" );
        done = FALSE;
	while( !done )
	{
		printf( "Transmit File specification: " );
		gets( TempName );
                if( index( TempName, ',' ) == NULL )
                       done = TRUE;
                  else *index( TempName, ',' ) = '\0';
                Fsetdta(&SysDta);
                if( Fsfirst(TempName, 0) == 0) {
                       do { list[ct].size_part = SysDta.fsize;
                            list[ct].time_part = SysDta.ftime;
                            list[ct].date_part = SysDta.fdate;
                            strcpy(list[ ct++ ].name_part, SysDta.fname);
                       } while (Fsnext() == 0);
                  }
                  else strcpy(list[ ct++ ].name_part, TempName);
                ct--;
	}

	batch = (ct ? TRUE : FALSE);
	for( i=0 ; i <= ct ; i++ )
	{
		while( rxstat() )	/* gobble garbage	*/
			rx();
		trans( list[ i ].name_part, list[i].size_part );/* send file*/
	}
	if( batch ) 
	{ 
		while( swait() != NAK ) /* await name request	*/
			abort();	/* operator abort scan	*/
		tx( ACK );		/* handshake NAK	*/
		sleep();		/* decent interval	*/
		tx( EOT );		/* end of transmission	*/
	}
}


/*---------------------------------------------------------------
trans() does transmission of a file.

References globals: buffer, fd, batch;
Modifies globals: fd, rec, crc;
---------------------------------------------------------------*/

VOID trans( in_name, in_size )
   char *in_name;
   long in_size;
{
   char name[ NAMESIZE+1];	/* expanded name		*/
   char ch;			/* scratch character		*/
   register char *i;		/* buffer pointer (index)	*/
   int y;			/* scratch counter		*/
   long records;
   
	rec = 1;		/* uses "natural" numbering	*/
	if( (fd = fopen( in_name, "br" )) == NULL)
	{
		printf( "Unable to open: %s\n", in_name );
		return;
	}
	if( batch )		/* need to send name?		*/
		txname( parse( in_name, name ) );
	printf( "\nTransmitting: %s %8ld bytes", in_name ,in_size);
	while( rxstat() )
		rx();

  records = in_size / 128;
  if ((records * 128) != in_size) records++;

	while( TRUE )		/* handshake w/ receiver	*/
	{
		printf("\nSyncronizing: " );
		if( (ch = waitcan( 10 )) == NAK )
		{
			printf( "Received Checksum Request\n" );
			crc = FALSE;
			break;
		}
		if( ch == CRC )
		{
			printf( "Received CRC Request\n" );
			crc = TRUE;
			break;
		}
	}
	while( fillbuf() )		/* file not empty?	*/
		txrec( buffer, records-- );/* send record	*/
	tx( EOT );
	while( (ch = waitcan( 1 )) != ACK )
	{
		tx( EOT );		/* show end		*/
	}
	fclose( fd );			/* dump file		*/
	printf( "\n%s transferred.", in_name );
}


/*---------------------------------------------------------------
txname() transmits a file name from the CP/M fcb format 
parameter 'name'.

References globals: buffer, checksum, crc;
Modifies globals: crc;
---------------------------------------------------------------*/

VOID txname( name )
   char *name;
{
   register int i;		/* scratch counter		*/
   char ch;			/* scratch char			*/
   char crcsav;			/* holds state of global 'crc'	*/

	crcsav = crc;		/* we always use checksum for	*/
	crc = FALSE;		/* name error detection		*/
	while( TRUE )		/* main loop for retries	*/
	{
		while( rxstat() )
			rx();
		i = 0;		/* retry count			*/
		while( TRUE )
		{
			printf( "\nWaiting for filename request");
			if( waitcan( RETRY ) == NAK )
				break;
			if( i++ > RETRY )
				error( "Can't send filename" );
			printf( " - Received garbage" );
		}
		tx( ACK );	/* handshake name request	*/
		sleep();	/* decent interval		*/
/*		printf( "\nSending name: " );                   */
		clrcrc();	/* clear checksum accumulator	*/
		for( i = 0 ; i < NAMESIZE ; i++ )/* name loop	*/
		{
			tx( name[ i ] );
/*			putchar( name[ i ] );	/* local echo	*/
			wait( 10 );	/* wait for ACK		*/
		}
		tx( EoF );		/* name terminated w/EoF*/
		if( wait( 10 ) == checksum )	/* handshake	*/
		{
/*			printf( " - Name sent OK" );            */
			tx( ACK );	/* handshake OK checksum*/
			crc = crcsav;	/* replacce 'crc'	*/
			return;		/* normal exit		*/
		}
		tx( BADNAME );		/* handshake bad chksm	*/
		printf( " - Bad name transfer" );
	}
}


/*---------------------------------------------------------------
txrec() transmits a single record, with retransmit on error from
receiver. Input is a pointer to the I/O buffer.

References globals: rec, crcaccum, checsum, crc;
Modifies globals: rec;
---------------------------------------------------------------*/

VOID txrec( buf, nrecs )
   char *buf;
long  nrecs;
{
   register int i;
   unsigned cr;

	while( TRUE )			/* do it until right	*/
	{
        	printf( "\rTransmitting record %7d %8ld ", rec, nrecs );
		tx( SOH );		/* start of header	*/
		tx( rec );		/* rec #		*/
		tx( ~rec );		/* 1's comp		*/
		clrcrc();		/* clear CRC accum	*/
		for( i = 0 ; i < RECSIZE ; i++ )
			tx( buf[ i ] );	/* send record		*/ 
		updcrc( 0 );		/* finish up CRC	*/
		updcrc( 0 );		/* again		*/
		cr = crcaccum; 		/* save crc lobyte	*/
		if( crc )		/* send hi byte first	*/
		{
			tx( crcaccum >> 8 );
			tx( cr );
		}
		else
			tx( checksum );
		if( waitcan(10)==ACK )	/* quit if correct	*/
			break;		
		printf( "Error in transmission\n" );
	}
	rec++;				/* bump record count	*/
}


/*---------------------------------------------------------------
receive() is the receive executive. It requires no input
parameters except 'crc' should be set. Does not "hunt" from 
CRC to checksum.

References globals: line;
Modifies globals: line;
---------------------------------------------------------------*/

VOID receive()
{
	printf( "\n\nReceive File specification: " );
	gets( line );
	if( rxstat() )
		rx();			/* gobble trash		*/
	if( index( line, '*' ) != NULL )/* 'batch' set by a 	*/
		rxbatch( line );	/* asterisk		*/
	else				/* in input file name	*/
		rxfile( line );		/* otherwise nobatch	*/
}


/*---------------------------------------------------------------
rxbatch() is the batch mode reception executive.

References Globals: line;
Modifies Globals: line;
---------------------------------------------------------------*/

VOID rxbatch()
{
   char fcb[ NAMESIZE + 1 ];	/* used by unparse()		*/
   char name[ NAMESIZE + 5];	/* unparsed name		*/

  while( TRUE ) {
       while( rxstat() ) rx();	/* gobble trash		*/
       printf( "\nWaiting for file name: " );
       rxname( fcb );	/* receive file name into fcb	*/
       unparse( name, fcb );	/* make "standard"	*/
       if( line[ 1 ] == ':' )
              line[ 2 ] = '\0';
         else line[ 0 ] = '\0';
       strcat( line, name );
       printf( "%s", line);
       rxfile( line );		/* receive file		*/
  }
}


/*---------------------------------------------------------------
rxname() loads a CP/M style fcb with a filename from remote
sender. On receipt of an EOT instead of a name character, 
aborts to master menu via longjmp().
---------------------------------------------------------------*/

VOID rxname( fcb )
   char *fcb;			/* points to CP/M style fcb	*/
{
   char *fcbptr;		/* index to fcb			*/
   register int i;		/* scratch counter		*/
   unsigned ch;			/* scratch char	(must hold ERR)	*/
   unsigned chksum;			/* checksum accumulator		*/

	while( TRUE )
	{
		fcbptr = fcb;		/* align index		*/
		i = RETRY * 5;		/* retry for name h.s.	*/
		while( TRUE )		/* Handshake NAK	*/
		{
			abort();	/* check operator abort	*/
			tx( NAK );	
			if( swait() == ACK )
				break;	
			if( !( i-- ) )
				error( "Timed out waiting for name" );
		}
		chksum = EoF;		/* init checksum	*/
		for( i = 0 ; i < 34 ; i++ )	/* accept noise	*/
		{
			if( (ch = swait()) == EOT )
			{
				tx( ACK );
				printf( 
				"\nAll transfers completed" );
				longjmp( &to_menu, 0 );
			}
			if( ch == EoF )	/* End of name chars	*/
				break;
			if( ch != ERROR )	/* not timeout	*/
			{
				*(fcbptr++) = ch & 0x7F;
/*				putchar( ch ); moved to rxbatch */
				chksum += ch & 0x7F;
			}
			abort();	/* operator abort	*/
			tx( ACK );	/* handshake name char	*/
		}
		fcb[ NAMESIZE ] = 0;	/* terminate name field	*/
		do {
			abort();	/* operator abort	*/
			tx( chksum );	/* handshake checksum	*/
		} while( (ch = swait()) == ERROR );
		if( ch == ACK )		/* ACK is good name	*/
			return;
		printf(" - Checksum error\nRetrying file name: " );
	}
}

	
/*---------------------------------------------------------------
rxfile() receives a file. Input parameter is the name, including
drivspec if any, to save incoming data under. To maintain speed
the function is rather long to reduce function calls where
possible. Have 'crc' set before entry.

References globals: buffer, fd, rec, crcaccum, checksum, crc;
Modifies globals: buffer, rec;
---------------------------------------------------------------*/

VOID rxfile( name )
   char *name;
{
   char ch;			/* scratch handshake var	*/
   char response;		/* ACK/NAK/CRC handshake	*/
   unsigned crcrlo;		/* rec'd CRC low byte		*/
   unsigned crcrhi;		/* rec'd CRC hi byte		*/
   char r1;			/* current record number	*/
   char r2;			/* 1's comp record number	*/
   unsigned int j;		/* wait loop timer		*/
   int i;			/* scratch counter		*/
   register char *bptr;		/* buffer index			*/

	while( rxstat() )
		rx();
	make( name );		/* open/make backup		*/
	printf( "\n" );
	rec = 1;		/* uses natural numbering	*/
	if( crc )		/* set initial handshake to	*/
		response = CRC;	/* CRC or checksum		*/
	else
		response = NAK;
	while( TRUE )		/* record receive loop		*/
	{
		bptr = buffer;		/* align index		*/
		printf( "\rWaiting for record %7d %7dk ", rec, (rec / 8) );
		for( i=1 ; i <= RETRY * 5 ; i++ )
		{
			abort();
			tx( response );	/* send handshake	*/
			if( (ch = swait()) == SOH )
				break;	/* SOH indicatees rec	*/
			if( ch == EOT )	/* EOT indicates done	*/
			{
				printf( "\n%s received OK", name );
				fclose( fd );
				tx( ACK );	/* handshake 	*/
				return;		/* normal exit	*/
			}
			if( ch == CAN )	/* Xmit request abort	*/
			{
				fclose( fd );
				error("\rReceived cancel request");
			}
			if( i == RETRY * 5 )	/* timeout exit	*/
				error( "Can't sync to sender" );
		}
		r1 = swait();		/* record number	*/
		r2 = swait();		/* 1's comp record #	*/
		while( bptr - buffer < RECSIZE ) /* test count	*/
		{
			*(bptr++) = (char)swait();	/* accept char	*/
		}
		if( crc )			/* get hibyte CRC	*/
			crcrhi = swait() & 0xff;
		crcrlo = swait() & 0xff;	/* lobyte CRC or chksm	*/
		response = NAK;		/* init response	*/
		if( (~r1 & 0xFF) != (r2 & 0xFF) )
		{
			printf( "- Record numbers don't match\n" );
			continue;
		}
		clrcrc();		/* calc checksum/CRC	*/
		for( j = 0 ; j < RECSIZE ; j++ )
			updcrc( buffer[ j ] );
		updcrc( 0 );		/* required to finish	*/
		updcrc( 0 );		/* off CRC - why?	*/
		if( crc )		/* CRC test		*/
			if( (crcrlo | (crcrhi << 8)) != crcaccum )
			{
				printf( "- Bad CRC received\n" );
				continue;
			}
		if( !crc )		/* checksum test	*/
			if( (crcrlo & 0xff) != (checksum & 0xff) )
			{
				printf("- Bad checksum received\n");
				continue;
			}
		if( (r1 == (rec - 1) & 0xFF ) )	/* duplicate?	*/
		{
		printf( " - Received duplicate record\n" );
			response = ACK;	/* dup is OK - ACK was	*/
			continue;	/* trashed - ignore it	*/
		}
		if( r1 != (rec & 0xFF) )/* fatal sequence error	*/
			error( "File record numbering error" );
		rec++;			/* bump record count	*/
		for( j = 0 ; j < RECSIZE ; j++ ) /* write data	*/	
			putc( buffer[ j ], fd );
		response = ACK;		/* normal loop end	*/
	}
}


/*---------------------------------------------------------------
make() opens a file for output. It renames any existing file 
with the same name to ".BAK", and erases any previous backup.
Input parameter is filespec in "normal" (compressed) format.
Note: file rename by the link() function isn't supported
by several CP/M C Compilers, and others call this function
rename(), or reverse the arguments. At any rate, you can determine
how to handle backups in your own system and take any
appropriate action.

References globals: fd;
Modifies globals: fd;
---------------------------------------------------------------*/

VOID make( name )
   char *name;
{
#ifdef LINK
   char bak[ NAMESIZE + 5 ];	/* holds .BAK filespec		*/

				/* does this name exist?	*/
	if( (fd = fopen( name, "r" )) != NULL )
	{
		fclose( fd );	/* we were just checking	*/
		strcpy( bak, name );	/* save old name	*/
		if( index( name, '.' ) )/* extension spec'ed?	*/
			strcpy( index( bak, '.' ), ".BAK" );
		else
			strcat( bak, ".BAK" );
		unlink( bak );	/* dump any old .BAK file	*/
/* 		link( bak, name ); /* rename new .BAK file	*/
                Frename(0, name, bak);/* rename name -> bak     */
	}
#endif
	if( (fd = fopen( name, "bw" )) == NULL )
		error( "Can't create file - Aborting" );
}


/*---------------------------------------------------------------
parse() expands  non - ambiguous filespecs to the
standard CP/M fcb format, excluding drive byte. 
Sets the globals
'batch' if ambigous, and 'driv' for drivespecs.
Inputs are "normal" filespec (inspec), and exanded fcb.
---------------------------------------------------------------*/

char *parse( inspec, fcb )
   char *inspec, *fcb;
{
   register int i;			/* fcb index		*/
   int inptr;				/* input spec index	*/

	for( i = 0 ; i < NAMESIZE ; i++ )/* blank fill name	*/
		fcb[ i ] = ' ';
	if( inspec[ 1 ] == ':' )	/* check for drivspec	*/
		inptr = 2;		/* point past drivespec	*/
	else
		inptr = 0;		/* index to start	*/
	i = 0;				/* pointer into fcb	*/
	while( TRUE )
		switch( inspec[ inptr++ ] )
		{
			case '\0':	/* end of input spec	*/
				fcb[ NAMESIZE ] = 0;
				return( fcb );
			case '.':	/* extension spec'ed	*/
				i = NAMESIZE - 3; /* extension	*/
				break;
			default:
				if( i < NAMESIZE )
					fcb[ i++ ] = toupper( 
					inspec[ inptr - 1 ] );
		}
}


/*---------------------------------------------------------------
fillbuf() loads the I/O buffer with a record from the input file.
Returns TRUE if data was available, FALSE if no data left.

References globals: buffer, fd;
Modifies globals: buffer;
---------------------------------------------------------------*/

int fillbuf()
{
   register int i;		/* scratch counter		*/
   int errorchk;		/* holds EOF in Eco C		*/

	for( i = 0 ; i < RECSIZE ; i++ )
	{
		if( (errorchk = getc( fd)) == EOF )
			break;
		buffer[ i ] = errorchk;
	}
	if( i == 0 )			/* no data read		*/
		return( FALSE );
	for( ; i < RECSIZE ; i++ )
		buffer[ i ] = 0;	/* zero fill at EOF	*/
	return( TRUE );
}


/*---------------------------------------------------------------
unparse() reassembles a filename from a CP/M fcb into "normal" or
cmpressed form, so that our C functions can deal with them.
Inputs are a pointer to a string to receive the name (name), and
a pointer to a CP/M style fcb entry (buf).
---------------------------------------------------------------*/

char *unparse( name, buf )
   char *name, *buf;
{
   register int i;		/* 'name' index			*/
   int j;			/* 'buf' index			*/

	i = 0;			/* 'driveless' name	*/
	for( j = 0 ; j < NAMESIZE ; j++ )/* transfer chars	*/
	{
		if( buf[ j ] != ' ' )	/* (skip spaces)	*/
			name[ i++ ] = buf[ j ] & 0x7F;
		if( j == NAMESIZE-4 )	/* don't forget dot	*/
			name[ i++ ] = '.';
	}
	name[ i ] = '\0';		/* terminate string	*/
					/* eat terminal dot	*/
	if( *(index( name, '.' ) + 1) == '\0' )
		*(index( name, '.' )) = '\0'; 
	return( name );			/* return pointer	*/
}


/*---------------------------------------------------------------
clrcrc() clears the crc accumulator. Not much to it, actually.

References globals: 
Modifies globals: crcaccum, checksum;
---------------------------------------------------------------*/

VOID clrcrc()
{
	crcaccum = 
	checksum = 0;
}

/*---------------------------------------------------------------
updcrc() updates the crc accumulator, if 'crc' is TRUE, else
updates the checksum. 
'x' is the byte to be added to CRC or checksum.
CCITT polynomial.

References globals: crc;
Modifies globals: crcaccum, checksum;
---------------------------------------------------------------*/

VOID updcrc( x )
   char x;
{
   unsigned shifter, i, flag;

	if( crc )
	{
		for( shifter = 0x80 ; shifter ; shifter >>= 1 )
		{
			flag = (crcaccum & 0x8000);
			crcaccum <<= 1;
			crcaccum |= ((shifter & x) ? 1 : 0);
			if( flag )
				crcaccum ^= 0x1021;
		}
	}
	else
		checksum += x;
}


/*---------------------------------------------------------------
abort() scans console for ^X abort code. Exit via error().
System Dependent. If not supportable, convert to null function
or delete and use a preprocessor macro to make it go away.
---------------------------------------------------------------*/

VOID abort()
{
  if (Bconstat(2) != 0)
     if ((int)Bconin(2) == ('X'-'@') )
	error( "\rOperator requested abort" );
}


/*---------------------------------------------------------------
error() does error abort cleanup

References globals: fd;
---------------------------------------------------------------*/

VOID error( str )
   char *str;
{
	fclose( fd );			/* dump any open files	*/
	tx( CAN );			/* alert remote		*/
	putchar( BELL );		/* alert operator	*/
	printf("\n\nError - %s\nPress any key to continue", str );
	GetChar();			/* pause		*/
	longjmp( &to_menu, 0 );		/* rejoin menu		*/
}


/*---------------------------------------------------------------
sleep() does a short delay to account for transmission line
latency, etc.
---------------------------------------------------------------*/

VOID sleep()
{
   register unsigned int i;

	for( i=0 ; i < MAGIC_NUMBER ; i++ )
		;
}


/*---------------------------------------------------------------
wait() waits for a character - timeout built in. Timeout 
condition causes error return to menu.
Input parameter 'time' controls duration of wait.
---------------------------------------------------------------*/

unsigned wait( time )
   int time;
{
   register unsigned int i;		/* loop timer		*/
   int j;				/* timeout count	*/

	j =				/* timeout count	*/
	i = 0;				/* inner timeout count	*/
	while( !rxstat() )
		if( i++ > MAGIC_NUMBER * 3 )/* about 1.5 secs	*/
			if( !(time--) ) /* check retry count	*/
				error( "Receiver timed out" );
			else		/* more tries available	*/
			{
				abort(); /* scan for ^X		*/
				if( !j ) /* \n first timeout	*/
					printf( "\n" );
				printf( "\rtimeout %d   ", ++j );
				i = 0;
			}
	return( rx() );			/* send back char	*/
}


/*---------------------------------------------------------------
swait() does a short wait for a char from rx, returns ERROR on
a timeout, or char received.
---------------------------------------------------------------*/

unsigned swait()
{
   register unsigned int i;		/* loop timer		*/

	i = 0;
	while( i++ < (MAGIC_NUMBER * 6) )/* loop timing test	*/
		if( rxstat() )		/* anything yet?	*/
			return( rx() ); /* yes, return char	*/
	return( ERROR );		/* timeout exit		*/
}


/*---------------------------------------------------------------
waitcan() waits for a character - If recv's CAN aborts to main
menu. Uses wait() for timing loop, so timeout also returns
to main menu. Input argument 'time' controls duration of wait.
---------------------------------------------------------------*/

unsigned waitcan( time )
   int time;
{
   char ch;

	if( (ch = wait( time )) == CAN )
		error( "Received cancellation request" );
	return( ch );
}


/*---------------------------------------------------------------
=================================================================
The following functions are machine dependant and should be
modified to run on your machine.
=================================================================
---------------------------------------------------------------*/


/*---------------------------------------------------------------
init() initializes globals. In addition, UART initialization
goes here as well, if required. No input parameters. May have
some machine dependant stuff here. 

References globals:
Modifies globals: crc, batch, buffer, line;
---------------------------------------------------------------*/

VOID init()
{
  setbuf(stdin, NULL);
  setbuf(stdout, NULL);
  buffer = malloc( RECSIZE );	/* transfer buffer	*/
  line = malloc( 81 );		/* scatch input line	*/
  crc =				/* CRC mode		*/
  batch = TRUE;			/* not really required	*/
}


/*---------------------------------------------------------------
rx() receives a char from uart. No timeout checking is performed.
Also, no status check is done. Bad news if you call this without
checking status first... Done that way to speed up things.

                       ZORBA Version 
---------------------------------------------------------------*/

unsigned rx()
{
	return( (unsigned)Bconin(1) & 0x00ff );
}


/*---------------------------------------------------------------
rxstat() returns TRUE if char waiting for input.

              ZORBA Version (interrupt fifo)
---------------------------------------------------------------*/

int rxstat()
{
  if (Bconstat(1) == 0)
          return(FALSE);
     else return(TRUE);
}


/*---------------------------------------------------------------
tx( ch ) transmits the char 'ch', and calls the CRC/checksum 
update. A timeout is built in for transmission faults.

                      ZORBA Version
---------------------------------------------------------------*/

VOID tx( ch )
   char ch;
{

   register unsigned int i;		/* timing loop counter	*/

	i = 0;
	while( !txstat() )		/* ready?		*/
		if( i++ > MAGIC_NUMBER )/* loop about 3 secs	*/
			error( "UART not ready??" );
	Bconout(1, ch );		/* to UART		*/
	updcrc( ch );			/* stir CRC a little	*/
}


/*---------------------------------------------------------------
txstat() returns TRUE if tx register is empty

                     ZORBA Version
---------------------------------------------------------------*/

int txstat()
{
  if (Bcostat(1) == 0)
          return (FALSE);
     else return (TRUE);
}


/*---------------------------------------------------------------
quit() does exit processing.
---------------------------------------------------------------*/

VOID quit()
{
	exit( 0 );
}
